#pragma once

#include <iostream>
#include <SFML/System/Vector2.hpp>
#include <cmath>
#include <concepts>
#include <numbers>


#define DBG(X) ({ auto value = X; std::cout << "(DBG) [" << __FILE__ << ":" << __LINE__ << "] (" << #X << ") = " << value << "\n"; value;})


inline namespace utils {

    template <typename T>
    concept is_numeric= std::integral<T> || std::floating_point<T>;

    template<typename T> requires(is_numeric<T>)
    class RangeIterator {
        public:
            inline RangeIterator(T index, T increment): _index(index), _increment(increment) {}
            const inline T operator*() { return _index; }
            const inline RangeIterator operator++() { _index += _increment; return *this; }
            // note : incorrect comparison needed to support saturating iterations.
            inline friend bool operator== (const RangeIterator<T>& start, const RangeIterator<T>& end) { return start._index >= end._index; }
            inline friend bool operator!= (const RangeIterator<T>& start, const RangeIterator<T>& end) { return !(start._index >= end._index); }
        private:
            T _index;
            T _increment;
    };

    template<typename T> requires(is_numeric<T>)
    class Range {
        public:
            inline Range(T to) : _from(0), _to(to), _increment(1) {};
            inline Range(T from, T to) : _from(from), _to(to), _increment(1) {};
            inline Range(T from, T to, T increment) : _from(from), _to(to), _increment(increment) {};
            const inline RangeIterator<T> begin() {
                return RangeIterator(_from, _increment);
            }
            const inline RangeIterator<T> end() {
                return RangeIterator(_to, _increment);
            }
        private:
            T _from;
            T _to;
            T _increment;
    };


    template<typename T> requires(is_numeric<T>)
    inline T vectorLength(const sf::Vector2<T>& v) {
        return std::sqrt(v.x * v.x + v.y * v.y);
    }


    inline float deg_to_rad(float deg) { return (deg / 360) * (2 * std::numbers::pi); }
    inline float rad_to_deg(float rad) {
      return (rad / (2 * std::numbers::pi)) * (360);
    }

    inline sf::Vector2f scaled(sf::Vector2f vec, float fac) { return sf::Vector2f(vec.x * fac, vec.y * fac); }
}
